#Import "<std>"
#Import "<mojo>"
#Import "B2D/mx2_box2d/b2Manager"

#Import "assets/d1a.json"
#Import "assets/rorov77.json"
#Import "assets/images/@/images/"

Using std..
Using mojo..
Using box2d..

Global w_width:=1200 'initial window size
Global w_height:=760

Class Box2DgfxTest Extends Window

	Field physManager:b2Manager 'the box2d Manager
	
	Field desertBackroundImage:=New Image[4]

	Field scroll:Vec2f 'carema center viewpoint
	Field zoom:=1.0 'camera zoom
	
	Field pscale:=61.0 'physics scale
	Field offset:=New Vec2f(100,0.3-1080/61.0) 'the offset for incoming json

	Field rovb2Joints:b2Joint[]
	
	Method New( title:String,width:Int,height:Int,flags:WindowFlags=WindowFlags.Resizable )
		
		Super.New( title,width,height,flags )
		ClearColor=Color.Black
		
		' backround images (not physics related)
		desertBackroundImage[0]=Image.Load("asset::images/d1.png")
		desertBackroundImage[1]=Image.Load("asset::images/d2.png")
		desertBackroundImage[2]=Image.Load("asset::images/d3.png")
		desertBackroundImage[3]=Image.Load("asset::images/d4.png")
		
		'------- Initialising b2Manager (the world and all the stuff associated wth the Json)
		physManager=New b2Manager("asset::d1a.json",pscale,True,offset) 'loading the "ground"
		physManager.AddJson("asset::rorov77.json",New b2Vec2(126,-6)) 'loading the "rover" 

		'getting the wheel joints called "rovWheel" in the json file (in the form of b2Joint Array), in order to control them
		rovb2Joints=physManager.GetJoints("rovWheel")
		
	End
	
	Method OnRender( canvas:Canvas ) Override
		
		App.RequestRender()
		
		canvas.PushMatrix()
		
			'centering the canvas transform on the rover
			scroll=physManager.FromPhysics(physManager.GetBody("rovL").GetPosition())
			scroll.y=-Height/2+1040 'fixing the height at constant
			canvas.SetCameraByCenter(scroll,zoom)
	
			' Drawing the backround (not physics related ), just a "manual" parallax - should be checking if visible before drawing it for better performances...
			For Local i:=0 Until 4
				For Local j:=0 To 4
				canvas.DrawImage(desertBackroundImage[i],scroll.x*(5-i)/7.0+j*1920,0,0,1.0,1.0)
				Next
			Next
			
			physManager.StepWorld() 'steping the physics world
			physManager.DrawBodies(canvas) 'drawing the images of the bodies as described in the json
			'physManager.DrawDebug(canvas)  'if you want the see the debug draw
		
		canvas.PopMatrix()
		
		'rover controls
		Local MotorSpeed:=0.0
		If Keyboard.KeyDown(Key.Right) Or (Mouse.ButtonDown(MouseButton.Left) And Mouse.Location.x > Width/2) Then MotorSpeed=-5.0
		If Keyboard.KeyDown(Key.Left) Or (Mouse.ButtonDown(MouseButton.Left) And Mouse.Location.x < Width/2 And Mouse.Location.x>0) Then MotorSpeed=5.0
		For Local i:=0 To 9
			If Touch.FingerDown( i ) And Touch.FingerLocation(i).x>Width/2 Then MotorSpeed=-5
			If Touch.FingerDown( i ) And Touch.FingerLocation(i).x<Width/2 Then MotorSpeed=5
		Next 
		For Local i:=0 Until 3 '3 wheels
			'you have to cast b2Joints to their actual type before using it (it is the box2d approach)
			'the mx2 box2d module offers casting methods, in this case .ToWheel(), trying to cast to the wrong type will generate crash..
			rovb2Joints[i].ToWheel().SetMotorSpeed(MotorSpeed) 
		Next
		
		'debug controls	
		If Keyboard.KeyPressed(Key.D)
			physManager.GetBody("rovL").ApplyForceToCenter(New b2Vec2 (0,500),true)
		End
		If Keyboard.KeyDown(Key.I)
			zoom*=1.01
		End
		If Keyboard.KeyDown(Key.O)
			zoom*=0.99
		End
		
		canvas.Color=Color.Black
		canvas.Scale(1.5,1.5)
		canvas.DrawText("Use left/right arrows or touch left/right screen zone to move. Backround from craftpix.net. FPS: "+App.FPS,5,15)
		canvas.DrawText("Mouse Location: "+Mouse.Location,5,30)
		
	End
End

Function Main()
	New AppInstance
	New Box2DgfxTest( "Hopa_Box2D_test",w_width,w_height )
	App.Run()
End

Class Canvas Extension
	Method SetCameraByCenter(point:Vec2f,zoom:Float=1.0,rotation:Float=0)
		Translate(Viewport.Width/2,Viewport.Height/2)
		Scale(zoom,zoom)
		Rotate(rotation)
		Translate(-point)
	End
End
